
#include<cstdlib>
#include<string>
#include<iostream>

#include "CACHE.hpp"

#ifndef _H_CS_MATRIX
#define _H_CS_MATRIX

/** Denotes the base simulated datatype. */
template < typename T > class CS_MATRIX {
   private:
	/** Do not use standard constructor. */
	CS_MATRIX() {}

   protected:
	/** Contains the actual data. */
        T** _matrix;

	/** column size. */
	unsigned long int _cols;

	/** row size. */
	unsigned long int _rows;

   public:

	//static stuff first

	static std::string fDim( const CS_MATRIX& obj ) {
		std::ostringstream oss;
		oss << "( " << obj.rows() << " x " << obj.cols() << " )";
		return oss.str();
	}

	//end static stuff


	/** Description of the current variable. */
	std::string _descr;

	/** The only valid constructor. */
	CS_MATRIX( unsigned long int rows, unsigned long int cols, std::string descr ) {
		if ( descr == "" ) {
			std::cerr << "Parameter description is mandatory during CS_MATRIX initialisation!" << std::endl;
			exit( 1 );
		}

		if ( cols == 0 || rows == 0 ) {
			std::cerr << "Size cannot be 0 during allocation of " << descr << std::endl;
			exit( 1 );
		}

		typedef T* T_POINTER;
		_matrix = new T_POINTER[ rows ];
		for ( unsigned long int i=0; i<rows; i++ )
			_matrix[ i ] = new T[ cols ];
		_cols = cols;
		_rows = rows;
		_descr = descr;
	}

	/** Returns the size of the array. */
	unsigned long int cols() const { return _cols; }

	/** Returns the size of the array. */
	unsigned long int rows() const { return _rows; }

	/** Matrix accessor */
	inline
	T& operator() ( unsigned long int i, unsigned long int j ) {
		return access( i, j );
	}

	/** Matrix access, function version. */
	T& access( unsigned long int i, unsigned long int j ) {
		if ( i < _rows && j < _cols ) {
			CACHE::getInstance()->access( &( _matrix[ i ][ j ] ), sizeof( T ) );
			return _matrix[ i ][ j ];
		} else {
			std::cerr << _descr << ": Index out of bounds! (" << i << "," << j << ") is not in [0," << _rows << "] x [0," << _cols << "] !" << std::endl;
			exit( !EXIT_SUCCESS );
		}
	}

	/** Const matrix access version. */
	const T& const_access( unsigned long int i, unsigned long int j ) const {
		if ( i < _rows && j < _cols ) {
			CACHE::getInstance()->access( &( _matrix[ i ][ j ] ), sizeof( T ) );
			return _matrix[ i ][ j ];
		} else {
			std::cerr << _descr << ": Index out of bounds! (" << i << "," << j << ") is not in [0," << _rows << "] x [0," << _cols << "] !" << std::endl;
			exit( !EXIT_SUCCESS );
		}
	}

	~CS_MATRIX() {
		for ( unsigned int i = 0; i < _rows; i++ )
			delete [] _matrix[ i ];
		delete [] _matrix;
	}

};

#endif

